
{%- if pagename != "search" %}
<div id="searchbox" style="display: none">
    <form class="search" action="{{ pathto('search') }}" method="get">
      <input type="text" name="q" size="18" />
      <input type="submit" value="{{ _('Search') }}" />
      <input type="hidden" name="check_keywords" value="yes" />
      <input type="hidden" name="area" value="default" />
    </form>
</div>
<script type="text/javascript">$('#searchbox').show(0);</script>
{%- endif %}

<h3>quicklinks</h3>
<div style="text-align: left; font-size: 100%; vertical-align: middle;">
<table>
<tr>
<td>
        <a href="{{ pathto('index') }}">home</a>
</td><td>
        <a href="{{ pathto('examples') }}">examples</a>
</td></tr><tr><td>
        <a href="{{ pathto('install') }}">install</a>
</td><td>
        <a href="{{ pathto('changelog') }}">changelog</a>
</td></tr><tr><td>
        <a href="{{ pathto('config') }}">config</a>
</td><td>
        <a href="https://github.com/tox-dev/tox/issues">issues[gh]</a>
</td></tr><tr><td>
        <a href="{{ pathto('support') }}">support</a>
</td><td>
        <a href="{{ pathto('plugins') }}">plugins/hooks</a>
</td>
</tr></table>
</div>
{% extends "basic/localtoc.html" %}

